#<profile4xi profile = "10" />
# -------------------------------------------------------------------------------------------------
#<program4xi program = "_run_finalizers4xi.py">
#<abstract4xi abstract = "providing function _run_finalizers4xi"/>
#<refer4xi reference = "util.py"/>
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = ['_run_finalizers4xi'] 
#"
#/>

# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
import os
import sys
import threading                
import _thread
import _winapi   

import itertools
import weakref
import atexit
import logging
#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
import _multiprocessing
#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
import finalizer4xi
from finalizer4xi import clear_finalizer_registry4xi
from print4xi import print4xi
from time4xi import time4xi
#"
#/>




# -------------------------------------------------------------------------------------------------
#<function4xi function = "run-finalizers4xi" usage = "
# _exit_function():
#     _run_finalizers(0)<--- min priority 0
#     _run_finalizers() <--- min priority None
#"> 
#<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
# -------------------------------------------------------------------------------------------------
def _run_finalizers4xi(minpriority=None):
    '''
    runs all finalizers whose exit priority is not None and at least minpriority

    - finalizers with highest priority are called first; finalizers with
    the same priority will be called in reverse order of creation
    '''
    
    # -----------------------------------------------------------------------------------------    
    # (Step 1/7)
    # -----------------------------------------------------------------------------------------        
        
    class2print = "_run_finalizers4xi"
    method2print = "_run_finalizers4xi"

    
    message2print = "{(Step 1/7a) started" + "; run-finalizers4xi-started: " + str(_run_finalizers4xi)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print) 
        
    message2print = "{(Step 1/7b) passed" + "; minimun-priority-passed-to-run-finalizers4xi: " + str(minpriority) + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)  
         
    
    # -----------------------------------------------------------------------------------------    
    # (Step 2/7)
    # -----------------------------------------------------------------------------------------        
       
    if minpriority is None:
        f = lambda p : p[0][0] is not None
        
        message2print = "{(Step 2/7a) set" + "; finalizer2xi-set-by-run-finalizers4xi to: " + str(f)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
                
    else:
        f = lambda p : p[0][0] is not None and p[0][0] >= minpriority
        
        message2print = "{(Step 2/7b) set" + "; finalizer2xi-set-by-run-finalizers4xi to: " + str(f)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
       
         
        
    # -----------------------------------------------------------------------------------------    
    # (Step 3/7)
    # -----------------------------------------------------------------------------------------    
    items = [x for x in list(finalizer4xi._finalizer_registry.items()) if f(x)]
    
    message2print = "{(Step 3/7) set" + "; items-set-by-run-finalizers4xi to: " + str(items)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)  
                
     
    # -----------------------------------------------------------------------------------------    
    # (Step 4/7)
    # -----------------------------------------------------------------------------------------        
    items.sort(reverse=True)
    
    message2print = "{(Step 4/7) performed" + "; items-sort-performed-by-run-finalizers4xi: " + str(items.sort) + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)             
               
     
    # -----------------------------------------------------------------------------------------    
    # (Step 5/7)
    # -----------------------------------------------------------------------------------------    
    for key, finalizer in items:
        # calling finalizer
        
        message2print = "{(Step 5/7a) got" + "; key-got-by-run-finalizers4xi: " + str(key)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                
                
        message2print = "{(Step 5/7b) calling" + "; finalizer-calling-by-run-finalizers4xi: " + str(finalizer) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)       
                
        
        try:
            finalizer()
            
            message2print = "{(Step 5/7c) performed" + "; finalizer-performed-by-run-finalizers4xi: " + str(finalizer) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)             
                    
                  
        except Exception:
            import traceback
            traceback.print_exc()
            
            message2print = "{(Step 5/7d) hit" + "; key-hit-by-run-finalizers4xi: " + " *** " + str(key)  + " *** " + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
        
            
            message2print = "{(Step 5/7e hit" + "; finalizer-hit-by-run-finalizers4xi: " + " *** " + str(finalizer)  + " *** " + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
                
         
             
    # -----------------------------------------------------------------------------------------    
    # (Step 6/7)
    # -----------------------------------------------------------------------------------------              
    if minpriority is None:
    
        message2print = "{(Step 6/7a) calling" + "; clear-finalizer-registry4xi-calling-by-run-finalizers4xi: " + str(clear_finalizer_registry4xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)       
            
        clear_finalizer_registry4xi()
          
        message2print = "{(Step 6/7b) performed" + "; clear-finalizer-registry4xi-performed-by-run-finalizers4xi: " + str(clear_finalizer_registry4xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
             
    
        
    # -----------------------------------------------------------------------------------------    
    # (Step 7/7)
    # -----------------------------------------------------------------------------------------                   
    message2print = "{(Step 7/7) returning" + "; run-finalizers4xi-returning: " + time4xi() + "}" 
    print4xi(class2print, method2print, message2print)         
#</function4xi>

#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   test_run_finalizers4xi.txt
#   sense-test-run-finalizers4xi.xml
#   reason-test-run-finalizers4xi.xml
#   ace-test-run-finalizers4xi.xml
#"
#>

# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 

#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of _run_finalizers4xi.py                             
# ------------------------------------------------------------------------------------------------- 
#"
#/>
  